/// <reference types="node" />
import { TwingTokenParserInterface } from "./token-parser-interface";
import { TwingNodeVisitorInterface } from "./node-visitor-interface";
import { TwingExtensionCore } from "./extension/core";
import { TwingExtensionInterface } from "./extension-interface";
import { TwingFilter } from "./filter";
import { TwingLexer } from "./lexer";
import { TwingParser } from "./parser";
import { TwingTokenStream } from "./token-stream";
import { TwingSource } from "./source";
import { TwingLoaderInterface } from "./loader-interface";
import { TwingTest } from "./test";
import { TwingFunction } from "./function";
import { TwingTemplate } from "./template";
import { TwingTemplateWrapper } from "./template-wrapper";
import { TwingCacheInterface } from "./cache-interface";
import { TwingNode } from "./node";
import { TwingNodeModule } from "./node/module";
import { TwingRuntimeLoaderInterface } from "./runtime-loader-interface";
import { TwingOperator } from "./extension";
import { EventEmitter } from 'events';
import { SourceMapGenerator } from "source-map";
/**
 *  * Available options:
 *
 *  * debug: When set to true, it automatically set "auto_reload" to true as
 *           well (default to false).
 *
 *  * charset: The charset used by the templates (default to UTF-8).
 *
 *  * base_template_class: The base template class to use for generated
 *                         templates (default to Twig_Template).
 *
 *  * cache: An absolute path where to store the compiled templates,
 *           a TwingCacheInterface implementation,
 *           or false to disable compilation cache (default).
 *
 *  * auto_reload: Whether to reload the template if the original source changed.
 *                 If you don't provide the auto_reload option, it will be
 *                 determined automatically based on the debug value.
 *
 *  * strict_variables: Whether to ignore invalid variables in templates
 *                      (default to false).
 *
 *  * autoescape: Whether to enable auto-escaping (default to html):
 *                  * false: disable auto-escaping
 *                  * html, js: set the autoescaping to one of the supported strategies
 *                  * name: set the autoescaping strategy based on the template name extension
 *                  * PHP callback: a PHP callback that returns an escaping strategy based on the template "name"
 *
 *  * optimizations: A flag that indicates which optimizations to apply
 *                   (default to -1 which means that all optimizations are enabled;
 *                   set it to 0 to disable).
 */
export declare type TwingEnvironmentOptions = {
    debug?: boolean;
    charset?: string;
    base_template_class?: string;
    cache?: TwingCacheInterface | string | false;
    auto_reload?: boolean;
    strict_variables?: boolean;
    autoescape?: string | boolean | Function;
    optimizations?: number;
    source_map?: boolean | string;
};
/**
 * Stores the Twing configuration.
 *
 * @author Eric MORAND <eric.morand@gmail.com>
 */
export declare class TwingEnvironment extends EventEmitter {
    private charset;
    private loader;
    private debug;
    private autoReload;
    private cache;
    private lexer;
    private parser;
    private baseTemplateClass;
    private globals;
    private resolvedGlobals;
    private loadedTemplates;
    private strictVariables;
    private templateClassPrefix;
    private originalCache;
    private extensionSet;
    private runtimeLoaders;
    private runtimes;
    private optionsHash;
    private loading;
    private sourceMapNode;
    private sourceMap;
    /**
     * Constructor.
     *
     * @param {TwingLoaderInterface} loader
     * @param {TwingEnvironmentOptions} options
     */
    constructor(loader: TwingLoaderInterface, options?: TwingEnvironmentOptions);
    /**
     *
     * @returns {TwingExtensionCore}
     */
    getCoreExtension(): TwingExtensionCore;
    /**
     * Gets the base template class for compiled templates.
     *
     * @returns {string} The base template class name
     */
    getBaseTemplateClass(): string;
    /**
     * Sets the base template class for compiled templates.
     *
     * @param {string} templateClass The base template class name
     */
    setBaseTemplateClass(templateClass: string): void;
    /**
     * Enables debugging mode.
     */
    enableDebug(): void;
    /**
     * Disables debugging mode.
     */
    disableDebug(): void;
    /**
     * Checks if debug mode is enabled.
     *
     * @returns {boolean} true if debug mode is enabled, false otherwise
     */
    isDebug(): boolean;
    /**
     * Enables the auto_reload option.
     */
    enableAutoReload(): void;
    /**
     * Disables the auto_reload option.
     */
    disableAutoReload(): void;
    /**
     * Checks if the auto_reload option is enabled.
     *
     * @returns {boolean} true if auto_reload is enabled, false otherwise
     */
    isAutoReload(): boolean;
    /**
     * Enables the strict_variables option.
     */
    enableStrictVariables(): void;
    /**
     * Disables the strict_variables option.
     */
    disableStrictVariables(): void;
    /**
     * Checks if the strict_variables option is enabled.
     *
     * @returns {boolean} true if strict_variables is enabled, false otherwise
     */
    isStrictVariables(): boolean;
    /**
     * Gets the active cache implementation.
     *
     * @param {boolean} original Whether to return the original cache option or the real cache instance
     *
     * @returns {TwingCacheInterface|string|false} A TwingCacheInterface implementation,
     *                                                an absolute path to the compiled templates,
     *                                                or false to disable cache
     */
    getCache(original?: boolean): TwingCacheInterface | string | false;
    /**
     * Sets the active cache implementation.
     *
     * @param {TwingCacheInterface|string|false} cache A TwingCacheInterface implementation,
     *                                                an absolute path to the compiled templates,
     *                                                or false to disable cache
     */
    setCache(cache: TwingCacheInterface | string | false): void;
    /**
     * Gets the template class associated with the given string.
     *
     * The generated template class is based on the following parameters:
     *
     *  * The cache key for the given template;
     *  * The currently enabled extensions;
     *  * Whether the Twig C extension is available or not;
     *  * PHP version;
     *  * Twig version;
     *  * Options with what environment was created.
     *
     * @param {string} name The name for which to calculate the template class name
     * @param {number|null} index The index if it is an embedded template
     *
     * @return string The template class name
     */
    getTemplateClass(name: string, index?: number): string;
    /**
     * Checks if the source_map option is enabled.
     *
     * @returns {boolean} true if source_map is enabled, false otherwise
     */
    isSourceMap(): string | boolean;
    /**
     * Renders a template.
     *
     * @param {string} name The template name
     * @param {{}} context An array of parameters to pass to the template
     * @returns {Promise<string>}
     */
    render(name: string, context?: any): string;
    /**
     * Displays a template.
     *
     * @param {string} name The template name
     * @param {{}} context An array of parameters to pass to the template
     *
     * @throws TwingErrorLoader  When the template cannot be found
     * @throws TwingErrorSyntax  When an error occurred during compilation
     * @throws TwingErrorRuntime When an error occurred during rendering
     */
    display(name: string, context?: any): void;
    /**
     * Loads a template.
     *
     * @param {string | TwingTemplateWrapper | TwingTemplate} name The template name
     *
     * @throws {TwingErrorLoader}  When the template cannot be found
     * @throws {TwingErrorRuntime} When a previously generated cache is corrupted
     * @throws {TwingErrorSyntax}  When an error occurred during compilation
     *
     * @returns {TwingTemplateWrapper}
     */
    load(name: string | TwingTemplateWrapper | TwingTemplate): TwingTemplateWrapper;
    /**
     * Loads a template internal representation.
     *
     * This method is for internal use only and should never be called
     * directly.
     *
     * @param {string} name  The template name
     * @param {number} index The index if it is an embedded template
     *
     * @returns {TwingTemplate} A template instance representing the given template name
     *
     * @throws {TwingErrorLoader}  When the template cannot be found
     * @throws {TwingErrorRuntime} When a previously generated cache is corrupted
     * @throws {TwingErrorSyntax}  When an error occurred during compilation
     *
     * @internal
     */
    loadTemplate(name: string, index?: number): TwingTemplate;
    /**
     * Creates a template from source.
     *
     * This method should not be used as a generic way to load templates.
     *
     * @param {string} template The template name
     *
     * @returns {TwingTemplate} A template instance representing the given template name
     *
     * @throws TwingErrorLoader When the template cannot be found
     * @throws TwingErrorSyntax When an error occurred during compilation
     */
    createTemplate(template: string): TwingTemplate;
    /**
     * Returns true if the template is still fresh.
     *
     * @param {string} name The template name
     * @param {number} time The last modification time of the cached template
     *
     * @returns {boolean} true if the template is fresh, false otherwise
     */
    isTemplateFresh(name: string, time: number): boolean;
    /**
     * Tries to load a template consecutively from an array.
     *
     * Similar to loadTemplate() but it also accepts instances of TwingTemplate and
     * TwingTemplateWrapper, and an array of templates where each is tried to be loaded.
     *
     * @param {string|TwingTemplate|TwingTemplateWrapper|Array<string|TwingTemplate>} names A template or an array of templates to try consecutively
     *
     * @returns {TwingTemplate|TwingTemplateWrapper}
     *
     * @throws {TwingErrorLoader} When none of the templates can be found
     * @throws {TwingErrorSyntax} When an error occurred during compilation
     */
    resolveTemplate(names: string | TwingTemplate | TwingTemplateWrapper | Array<string | TwingTemplate>): TwingTemplate | TwingTemplateWrapper;
    setLexer(lexer: TwingLexer): void;
    /**
     * Tokenizes a source code.
     *
     * @param {TwingSource} source The source to tokenize
     * @returns {TwingTokenStream}
     *
     * @throws {TwingErrorSyntax} When the code is syntactically wrong
     */
    tokenize(source: TwingSource): TwingTokenStream;
    setParser(parser: TwingParser): void;
    /**
     * Converts a token stream to a template.
     *
     * @param {TwingTokenStream} stream
     * @returns {TwingNodeModule}
     *
     * @throws {TwingErrorSyntax} When the token stream is syntactically or semantically wrong
     */
    parse(stream: TwingTokenStream): TwingNodeModule;
    /**
     * Compiles a node and returns its compiled templates.
     *
     * @returns { Map<number, TwingTemplate> } The compiled templates
     */
    compile(node: TwingNode): string;
    /**
     *
     * @param {TwingSource} source
     * @returns {Map<number, TwingTemplate> }
     */
    compileSource(source: TwingSource): string;
    setLoader(loader: TwingLoaderInterface): void;
    /**
     * Gets the Loader instance.
     *
     * @return TwingLoaderInterface
     */
    getLoader(): TwingLoaderInterface;
    /**
     * Sets the default template charset.
     *
     * @param {string} charset The default charset
     */
    setCharset(charset: string): void;
    /**
     * Gets the default template charset.
     *
     * @return {string} The default charset
     */
    getCharset(): string;
    /**
     * Returns true if the given extension is registered.
     *
     * @param {string} name
     * @returns {boolean}
     */
    hasExtension(name: string): boolean;
    /**
     * Adds a runtime loader.
     */
    addRuntimeLoader(loader: TwingRuntimeLoaderInterface): void;
    /**
     * Gets an extension by name.
     *
     * @param {string} name
     * @returns {TwingExtensionInterface}
     */
    getExtension(name: string): TwingExtensionInterface;
    /**
     * Returns the runtime implementation of a Twig element (filter/function/test).
     *
     * @param {string} class_ A runtime class name
     *
     * @returns {{}} The runtime implementation
     *
     * @throws TwimgErrorRuntime When the template cannot be found
     */
    getRuntime(class_: string): any;
    /**
     *
     * @param {TwingExtensionInterface} extension
     * @param {string} name
     */
    addExtension(extension: TwingExtensionInterface, name?: string): void;
    /**
     * Registers an array of extensions.
     *
     * @param {Array<TwingExtensionInterface>} extensions An array of extensions
     */
    setExtensions(extensions: Array<TwingExtensionInterface>): void;
    /**
     * Returns all registered extensions.
     *
     * @returns Map<string, TwingExtensionInterface> A hash of extensions (keys are for internal usage only and should not be relied on)
     */
    getExtensions(): Map<string, TwingExtensionInterface>;
    addTokenParser(parser: TwingTokenParserInterface): void;
    /**
     * Gets the registered Token Parsers.
     *
     * @returns {Array<TwingTokenParserInterface>}
     *
     * @internal
     */
    getTokenParsers(): TwingTokenParserInterface[];
    /**
     * Gets registered tags.
     *
     * @return Map<string, TwingTokenParserInterface>
     *
     * @internal
     */
    getTags(): Map<string, TwingTokenParserInterface>;
    addNodeVisitor(visitor: TwingNodeVisitorInterface): void;
    /**
     * Gets the registered Node Visitors.
     *
     * @returns {Array<TwingNodeVisitorInterface>}
     *
     * @internal
     */
    getNodeVisitors(): TwingNodeVisitorInterface[];
    addFilter(filter: TwingFilter): void;
    /**
     * Get a filter by name.
     *
     * Subclasses may override this method and load filters differently;
     * so no list of filters is available.
     *
     * @param string name The filter name
     *
     * @return Twig_Filter|false A Twig_Filter instance or null if the filter does not exist
     *
     * @internal
     */
    getFilter(name: string): TwingFilter;
    registerUndefinedFilterCallback(callable: Function): void;
    /**
     * Gets the registered Filters.
     *
     * Be warned that this method cannot return filters defined with registerUndefinedFilterCallback.
     *
     * @return Twig_Filter[]
     *
     * @see registerUndefinedFilterCallback
     *
     * @internal
     */
    getFilters(): Map<string, TwingFilter>;
    /**
     * Registers a Test.
     *
     * @param {TwingTest} test
     */
    addTest(test: TwingTest): void;
    /**
     * Gets the registered Tests.
     *
     * @returns {Map<string, TwingTest>}
     */
    getTests(): Map<string, TwingTest>;
    /**
     * Gets a test by name.
     *
     * @param {string} name The test name
     * @returns {TwingTest} A TwingTest instance or null if the test does not exist
     */
    getTest(name: string): TwingTest;
    addFunction(aFunction: TwingFunction): void;
    /**
     * Get a function by name.
     *
     * Subclasses may override this method and load functions differently;
     * so no list of functions is available.
     *
     * @param {string} name function name
     *
     * @returns {TwingFunction} A TwingFunction instance or null if the function does not exist
     *
     * @internal
     */
    getFunction(name: string): TwingFunction;
    registerUndefinedFunctionCallback(callable: Function): void;
    /**
     * Gets registered functions.
     *
     * Be warned that this method cannot return functions defined with registerUndefinedFunctionCallback.
     *
     * @return Twig_Function[]
     *
     * @see registerUndefinedFunctionCallback
     *
     * @internal
     */
    getFunctions(): Map<string, TwingFunction>;
    /**
     * Registers a Global.
     *
     * New globals can be added before compiling or rendering a template;
     * but after, you can only update existing globals.
     *
     * @param {string} name The global name
     * @param {*} value The global value
     */
    addGlobal(name: string, value: any): void;
    /**
     * Gets the registered Globals.
     *
     * @return array An array of globals
     *
     * @internal
     */
    getGlobals(): Map<any, any>;
    /**
     * Merges a context with the defined globals.
     *
     * @param {Map<*, *>} context
     * @returns {Map<*, *>}
     */
    mergeGlobals(context: Map<any, any>): Map<any, any>;
    /**
     * Gets the registered unary Operators.
     *
     * @return Map<string, TwingOperator> A map of unary operators
     *
     * @internal
     */
    getUnaryOperators(): Map<string, TwingOperator>;
    /**
     * Gets the registered binary Operators.
     *
     * @return Map<string, TwingOperator> An array of binary operators
     *
     * @internal
     */
    getBinaryOperators(): Map<string, TwingOperator>;
    updateOptionsHash(): void;
    /**
     *
     * @param {number} line 0-based
     * @param {number} column 1-based
     * @param {string} name
     * @param {string} source
     */
    enterSourceMapBlock(line: number, column: number, name: string, source: string): void;
    leaveSourceMapBlock(): void;
    getSourceMap(): SourceMapGenerator;
}
