/// <reference types="node" />
import { TwingExtension, TwingOperator } from "../extension";
import { TwingTokenParserFor } from "../token-parser/for";
import { TwingTokenParserExtends } from "../token-parser/extends";
import { TwingTokenParserFrom } from "../token-parser/from";
import { TwingTokenParserMacro } from "../token-parser/macro";
import { TwingNode } from "../node";
import { TwingTokenParserIf } from "../token-parser/if";
import { TwingTokenParserSet } from "../token-parser/set";
import { TwingTokenParserBlock } from "../token-parser/block";
import { TwingTokenParserInclude } from "../token-parser/include";
import { TwingTokenParserWith } from "../token-parser/with";
import { TwingFunction } from "../function";
import { TwingTokenParserSpaceless } from "../token-parser/spaceless";
import { TwingTokenParserUse } from "../token-parser/use";
import { TwingTokenParserFilter } from "../token-parser/filter";
import { TwingTokenParserImport } from "../token-parser/import";
import { TwingTokenParserDo } from "../token-parser/do";
import { TwingTokenParserFlush } from "../token-parser/flush";
import { TwingTest } from "../test";
import { TwingFilter } from "../filter";
import { DateTime, Duration } from 'luxon';
import { TwingEnvironment } from "../environment";
import { TwingSource } from "../source";
export declare class TwingExtensionCore extends TwingExtension {
    private dateFormats;
    private numberFormat;
    private timezone;
    private escapers;
    /**
     * Defines a new escaper to be used via the escape filter.
     *
     * @param {string} strategy     The strategy name that should be used as a strategy in the escape call
     * @param {Function} callable   A valid PHP callable
     */
    setEscaper(strategy: string, callable: Function): void;
    /**
     * Gets all defined escapers.
     *
     * @returns {Map<string, Function>}
     */
    getEscapers(): Map<string, Function>;
    /**
     * Sets the default format to be used by the date filter.
     *
     * @param {string} format The default date format string
     * @param {string} dateIntervalFormat The default date interval format string
     */
    setDateFormat(format?: string, dateIntervalFormat?: string): void;
    /**
     * Gets the default format to be used by the date filter.
     *
     * @return array The default date format string and the default date interval format string
     */
    getDateFormat(): string[];
    /**
     * Sets the default timezone to be used by the date filter.
     *
     * @param {string} timezone The default timezone string or a TwingDateTimeZone object
     */
    setTimezone(timezone: string): void;
    /**
     * Gets the default timezone to be used by the date filter.
     *
     * @returns {string} The default timezone currently in use
     */
    getTimezone(): string;
    /**
     * Sets the default format to be used by the number_format filter.
     *
     * @param {number} decimal the number of decimal places to use
     * @param {string} decimalPoint the character(s) to use for the decimal point
     * @param {string} thousandSep  the character(s) to use for the thousands separator
     */
    setNumberFormat(decimal: number, decimalPoint: string, thousandSep: string): void;
    /**
     * Get the default format used by the number_format filter.
     *
     * @return array The arguments for number_format()
     */
    getNumberFormat(): (string | number)[];
    getTokenParsers(): (TwingTokenParserFor | TwingTokenParserExtends | TwingTokenParserFrom | TwingTokenParserMacro | TwingTokenParserIf | TwingTokenParserSet | TwingTokenParserBlock | TwingTokenParserInclude | TwingTokenParserWith | TwingTokenParserSpaceless | TwingTokenParserUse | TwingTokenParserFilter | TwingTokenParserImport | TwingTokenParserDo | TwingTokenParserFlush)[];
    getFilters(): Map<number, TwingFilter>;
    getFunctions(): Map<number, TwingFunction>;
    getTests(): Array<TwingTest>;
    getOperators(): [Map<string, TwingOperator>, Map<string, TwingOperator>];
}
/**
 * Cycles over a value.
 *
 * @param {Array} values
 * @param {number} position The cycle position
 *
 * @returns {string} The next value in the cycle
 */
export declare function twingCycle(values: Array<any>, position: number): any;
/**
 * Returns a random value depending on the supplied parameter type:
 * - a random item from a Traversable or array
 * - a random character from a string
 * - a random integer between 0 and the integer parameter.
 *
 * @param {TwingEnvironment} env
 * @param {*} values The values to pick a random item from
 *
 * @throws TwingErrorRuntime when values is an empty array (does not apply to an empty string which is returned as is)
 *
 * @returns {*} A random value from the given sequence
 */
export declare function twingRandom(env: TwingEnvironment, values?: any): any;
/**
 * Converts a date to the given format.
 *
 * <pre>
 *   {{ post.published_at|date("m/d/Y") }}
 * </pre>
 *
 * @param {TwingEnvironment} env
 * @param {DateTime|Duration|string} date A date
 * @param {string|null} format The target format, null to use the default
 * @param {string|null|false} timezone The target timezone, null to use the default, false to leave unchanged
 *
 * @return string The formatted date
 */
export declare function twingDateFormatFilter(env: TwingEnvironment, date: DateTime | Duration | string, format?: string, timezone?: string | null | false): string;
/**
 * Returns a new date object modified.
 *
 * <pre>
 *   {{ post.published_at|date_modify("-1day")|date("m/d/Y") }}
 * </pre>
 *
 * @param {TwingEnvironment} env
 * @param {DateTime|Duration|string} date A date
 * @param {string} modifier A modifier string
 *
 * @returns {DateTime} A new date object
 */
export declare function twingDateModifyFilter(env: TwingEnvironment, date: Date | DateTime | Duration | string, modifier: string): DateTime;
/**
 * Converts an input to a DateTime instance.
 *
 * <pre>
 *    {% if date(user.created_at) < date('+2days') %}
 *      {# do something #}
 *    {% endif %}
 * </pre>
 *
 * @param {TwingEnvironment} env
 * @param {Date | DateTime | Duration | number | string} date A date or null to use the current time
 * @param {string | null | false} timezone The target timezone, null to use the default, false to leave unchanged
 *
 * @returns {DateTime} A DateTime instance
 */
export declare function twingDateConverter(env: TwingEnvironment, date: Date | DateTime | Duration | number | string, timezone?: string | null | false): DateTime | Duration;
/**
 * Replaces strings within a string.
 *
 * @param {string} str  String to replace in
 * @param {Array<string>|Map<string, string>} from Replace values
 *
 * @returns {string}
 */
export declare function twingReplaceFilter(str: string, from: any): any;
/**
 * Rounds a number.
 *
 * @param value The value to round
 * @param {number} precision The rounding precision
 * @param {string} method The method to use for rounding
 *
 * @returns int|float The rounded number
 */
export declare function twingRound(value: any, precision?: number, method?: string): any;
/**
 * Number format filter.
 *
 * All of the formatting options can be left null, in that case the defaults will
 * be used.  Supplying any of the parameters will override the defaults set in the
 * environment object.
 *
 * @param {TwingEnvironment} env
 * @param {*} number A float/int/string of the number to format
 * @param {number} decimal the number of decimal points to display
 * @param {string} decimalPoint the character(s) to use for the decimal point
 * @param {string} thousandSep the character(s) to use for the thousands separator
 *
 * @returns string The formatted number
 */
export declare function twingNumberFormatFilter(env: TwingEnvironment, number: any, decimal: number, decimalPoint: string, thousandSep: string): any;
/**
 * URL encodes (RFC 3986) a string as a path segment or a hash as a query string.
 *
 * @param {string|{}} url A URL or a hash of query parameters
 *
 * @returns {string} The URL encoded value
 */
export declare function twingUrlencodeFilter(url: string | {}): string;
/**
 * Merges an array with another one.
 *
 * <pre>
 *  {% set items = { 'apple': 'fruit', 'orange': 'fruit' } %}
 *
 *  {% set items = items|merge({ 'peugeot': 'car' }) %}
 *
 *  {# items now contains { 'apple': 'fruit', 'orange': 'fruit', 'peugeot': 'car' } #}
 * </pre>
 *
 * @param {*} arr1 An array
 * @param {*} arr2 An array
 *
 * @return array The merged array
 */
export declare function twingArrayMerge(arr1: any, arr2: any): any[] | Map<any, any>;
/**
 * Slices a variable.
 *
 * @param {TwingEnvironment} env
 * @param item A variable
 * @param {number} start Start of the slice
 * @param {number} length Size of the slice
 * @param {boolean} preserveKeys Whether to preserve key or not (when the input is an object)
 *
 * @returns The sliced variable
 */
export declare function twingSlice(env: TwingEnvironment, item: any, start: number, length?: number, preserveKeys?: boolean): string | Map<any, any>;
/**
 * Returns the first element of the item.
 *
 * @param {TwingEnvironment} env
 * @param {*} item A variable
 *
 * @returns {*} The first element of the item
 */
export declare function twingFirst(env: TwingEnvironment, item: any): any;
/**
 * Returns the last element of the item.
 *
 * @param {TwingEnvironment} env
 * @param item A variable
 *
 * @returns The last element of the item
 */
export declare function twingLast(env: TwingEnvironment, item: any): any;
/**
 * Joins the values to a string.
 *
 * The separator between elements is an empty string per default, you can define it with the optional parameter.
 *
 * <pre>
 *  {{ [1, 2, 3]|join('|') }}
 *  {# returns 1|2|3 #}
 *
 *  {{ [1, 2, 3]|join }}
 *  {# returns 123 #}
 * </pre>
 *
 * @param {Array<*>} value An array
 * @param {string} glue The separator
 *
 * @returns {string} The concatenated string
 */
export declare function twingJoinFilter(value: Array<any>, glue?: string): string;
/**
 * Splits the string into an array.
 *
 * <pre>
 *  {{ "one,two,three"|split(',') }}
 *  {# returns [one, two, three] #}
 *
 *  {{ "one,two,three,four,five"|split(',', 3) }}
 *  {# returns [one, two, "three,four,five"] #}
 *
 *  {{ "123"|split('') }}
 *  {# returns [1, 2, 3] #}
 *
 *  {{ "aabbcc"|split('', 2) }}
 *  {# returns [aa, bb, cc] #}
 * </pre>
 *
 * @param {TwingEnvironment} env
 * @param {string} value A string
 * @param {string} delimiter The delimiter
 * @param {number} limit The limit
 *
 * @returns {Array<string>} The split string as an array
 */
export declare function twingSplitFilter(env: TwingEnvironment, value: string, delimiter: string, limit: number): any;
/**
 * @internal
 */
export declare function twingDefaultFilter(value: any, defaultValue?: any): any;
/**
 * Returns the keys for the given array.
 *
 * It is useful when you want to iterate over the keys of an array:
 *
 * <pre>
 *  {% for key in array|keys %}
 *      {# ... #}
 *  {% endfor %}
 * </pre>
 *
 * @param {Array<*>} array An array
 *
 * @returns {Array<*>} The keys
 */
export declare function twingGetArrayKeysFilter(array: Array<any>): any[];
/**
 * Reverses a variable.
 *
 * @param {TwingEnvironment} env
 * @param {string | Map<*, *>} item A traversable instance, or a string
 * @param {boolean} preserveKeys Whether to preserve key or not
 *
 * @returns The reversed input
 */
export declare function twingReverseFilter(env: TwingEnvironment, item: any, preserveKeys?: boolean): string | Map<any, any>;
/**
 * Sorts an array.
 *
 * @param {Array<*>} array
 *
 * @returns {Map<*,*>}
 */
export declare function twingSortFilter(array: Array<any>): Map<any, any>;
/**
 * @internal
 */
export declare function twingInFilter(value: any, compare: any): boolean;
/**
 * Returns a trimmed string.
 *
 * @returns {string}
 *
 * @throws TwingErrorRuntime When an invalid trimming side is used (not a string or not 'left', 'right', or 'both')
 */
export declare function twingTrimFilter(string: string, characterMask?: string, side?: string): any;
/**
 * Escapes a string.
 *
 * @param {TwingEnvironment} env
 * @param {*} string The value to be escaped
 * @param {string} strategy The escaping strategy
 * @param {string} charset The charset
 * @param {boolean} autoescape Whether the function is called by the auto-escaping feature (true) or by the developer (false)
 *
 * @returns {string}
 */
export declare function twingEscapeFilter(env: TwingEnvironment, string: any, strategy?: string, charset?: string, autoescape?: boolean): any;
/**
 * @internal
 */
export declare function twingEscapeFilterIsSafe(filterArgs: TwingNode): string[];
export declare function twingConvertEncoding(string: string, to: string, from: string): Buffer;
/**
 * Returns the length of a variable.
 *
 * @param {TwingEnvironment} env A TwingEnvironment instance
 * @param thing A variable
 *
 * @returns {number} The length of the value
 */
export declare function twingLengthFilter(env: TwingEnvironment, thing: any): any;
/**
 * Converts a string to uppercase.
 *
 * @param {TwingEnvironment} env
 * @param {string} string A string
 *
 * @returns {string} The uppercased string
 */
export declare function twingUpperFilter(env: TwingEnvironment, string: string): string;
/**
 * Converts a string to lowercase.
 *
 * @param {TwingEnvironment} env
 * @param {string} string A string
 *
 * @returns {string} The lowercased string
 */
export declare function twingLowerFilter(env: TwingEnvironment, string: string): string;
/**
 * Returns a titlecased string.
 *
 * @param {TwingEnvironment} env
 * @param {string} string A string
 *
 * @returns {string} The titlecased string
 */
export declare function twingTitleStringFilter(env: TwingEnvironment, string: string): any;
/**
 * Returns a capitalized string.
 *
 * @param {TwingEnvironment} env
 * @param {string} string A string
 *
 * @returns {string} The capitalized string
 */
export declare function twingCapitalizeStringFilter(env: TwingEnvironment, string: string): any;
/**
 * @internal
 */
export declare function twingEnsureTraversable(seq: any): any;
/**
 * Checks if a variable is empty.
 *
 * <pre>
 * {# evaluates to true if the foo variable is null, false, or the empty string #}
 * {% if foo is empty %}
 *     {# ... #}
 * {% endif %}
 * </pre>
 *
 * @param value A variable
 *
 * @returns {boolean} true if the value is empty, false otherwise
 */
export declare function twingTestEmpty(value: any): boolean;
/**
 * Checks if a variable is traversable.
 *
 * <pre>
 * {# evaluates to true if the foo variable is an array or a traversable object #}
 * {% if foo is iterable %}
 *     {# ... #}
 * {% endif %}
 * </pre>
 *
 * @param value A variable
 *
 * @return {boolean} true if the value is traversable
 */
export declare function twingTestIterable(value: any): boolean;
/**
 * Renders a template.
 *
 * @param {TwingEnvironment} env
 * @param {Map<*,*>} context
 * @param {string|Array<string>} template The template to render or an array of templates to try consecutively
 * @param variables The variables to pass to the template
 * @param {boolean} withContext
 * @param {boolean} ignoreMissing Whether to ignore missing templates or not
 * @param {boolean} sandboxed Whether to sandbox the template or not
 *
 * @returns {string} The rendered template
 */
export declare function twingInclude(env: TwingEnvironment, context: Map<any, any>, template: string | Array<string>, variables?: any, withContext?: boolean, ignoreMissing?: boolean, sandboxed?: boolean): string;
/**
 * Returns a template content without rendering it.
 *
 * @param {TwingEnvironment} env
 * @param {string} name The template name
 * @param {boolean} ignoreMissing Whether to ignore missing templates or not
 *
 * @return string The template source
 */
export declare function twingSource(env: TwingEnvironment, name: string, ignoreMissing?: boolean): string;
/**
 * Provides the ability to get constants from instances as well as class/global constants.
 *
 * Constant function makes no sense in JavaScript. To emulate the expected behavior, it is assumed that
 * so-called constants are keys of the TwingEnvironment.globals property.
 *
 * @param {TwingEnvironment} env The environment
 * @param {string} constant The name of the constant
 * @param object The object to get the constant from
 *
 * @returns {string}
 */
export declare function twingConstant(env: TwingEnvironment, constant: string, object: any): any;
/**
 * Batches item.
 *
 * @param {Array} items An array of items
 * @param {number} size  The size of the batch
 * @param fill A value used to fill missing items
 *
 * @returns Array<any>
 */
export declare function twingArrayBatch(items: Array<any>, size: number, fill?: any): Array<Map<any, any>>;
/**
 * Returns the attribute value for a given array/object.
 *
 * @param {TwingEnvironment} env
 * @param {TwingSource} source
 * @param {*} object The object or array from where to get the item
 * @param {*} item The item to get from the array or object
 * @param {Array<*>} _arguments An array of arguments to pass if the item is an object method
 * @param {string} type The type of attribute (@see Twig_Template constants)
 * @param {boolean} isDefinedTest Whether this is only a defined check
 * @param {boolean} ignoreStrictCheck Whether to ignore the strict attribute check or not
 * @param {boolean} sandboxed
 *
 * @returns mixed The attribute value, or a Boolean when $isDefinedTest is true, or null when the attribute is not set and $ignoreStrictCheck is true
 *
 * @throws TwingErrorRuntime if the attribute does not exist and Twig is running in strict mode and $isDefinedTest is false
 *
 * @internal
 */
export declare function twingGetAttribute(env: TwingEnvironment, source: TwingSource, object: any, item: any, _arguments?: Array<any>, type?: string, isDefinedTest?: boolean, ignoreStrictCheck?: boolean, sandboxed?: boolean): any;
export declare namespace twingGetAttribute {
    let cache: Map<string, Map<string, string>>;
}
