import { TwingSource } from "./source";
/**
 * Twing base error.
 *
 * This error class and its children must only be used when
 * an error occurs during the loading of a template, when a syntax error
 * is detected in a template, or when rendering a template. Other
 * errors must use regular JavaScript error classes (like when the template
 * cache directory is not writable for instance).
 *
 * To help debugging template issues, this class tracks the original template
 * name and line where the error occurred.
 *
 * Whenever possible, you must set these information (original template name
 * and line number) yourself by passing them to the constructor. If some or all
 * these information are not available from where you throw the exception, then
 * this class will guess them automatically (when the line number is set to -1
 * and/or the name is set to null). As this is a costly operation, this
 * can be disabled by passing false for both the name and the line number
 * when creating a new instance of this class.
 *
 * @author Fabien Potencier <fabien@symfony.com>
 */
export declare class TwingError extends Error {
    protected static registry: Map<string, string>;
    name_: string | Object;
    private lineno;
    private rawMessage;
    private sourcePath;
    private sourceCode;
    private previous;
    constructor(message: string, lineno?: number, source?: TwingSource | Object | null, previous?: Error);
    static register(key: string, value: string): void;
    getMessage(): string;
    getPrevious(): Error;
    /**
     * Gets the raw message.
     *
     * @return string The raw message
     */
    getRawMessage(): string;
    /**
     * Gets the template line where the error occurred.
     *
     * @return int The template line
     */
    getTemplateLine(): number | boolean;
    /**
     * Sets the template line where the error occurred.
     *
     * @param {number} lineno The template line
     */
    setTemplateLine(lineno: number | boolean): void;
    /**
     * Gets the source context of the Twig template where the error occurred.
     *
     * @return TwingSource|null
     */
    getSourceContext(): TwingSource;
    /**
     * Sets the source context of the Twig template where the error occurred.
     */
    setSourceContext(source?: TwingSource): void;
    guess(): void;
    appendMessage(rawMessage: string): void;
    private updateRepr;
    guessTemplateInfo(): void;
    protected init(lineno: number, source: TwingSource | Object | null): void;
}
