"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const loader_1 = require("../error/loader");
/**
 * Loads templates from other loaders.
 *
 * @author Eric MORAND <eric.morand@gmail.com>
 */
class TwingLoaderChain {
    /**
     * @param {Array<TwingLoaderInterface>} loaders
     */
    constructor(loaders = []) {
        this.hasSourceCache = new Map();
        this.loaders = [];
        this.TwingLoaderInterfaceImpl = this;
        for (let loader of loaders) {
            this.addLoader(loader);
        }
    }
    addLoader(loader) {
        this.loaders.push(loader);
        this.hasSourceCache = new Map();
    }
    getSourceContext(name) {
        let exceptions = [];
        for (let loader of this.loaders) {
            if (!loader.exists(name)) {
                continue;
            }
            try {
                return loader.getSourceContext(name);
            }
            catch (e) {
                if (e instanceof loader_1.TwingErrorLoader) {
                    exceptions.push(e.message);
                }
            }
        }
        throw new loader_1.TwingErrorLoader(`Template "${name}" is not defined${exceptions.length ? ' (' + exceptions.join(', ') + ')' : ''}.`);
    }
    exists(name) {
        if (this.hasSourceCache.has(name)) {
            return this.hasSourceCache.get(name);
        }
        for (let loader of this.loaders) {
            if (loader.exists(name)) {
                this.hasSourceCache.set(name, true);
                return true;
            }
        }
        this.hasSourceCache.set(name, false);
        return false;
    }
    getCacheKey(name) {
        let exceptions = [];
        for (let loader of this.loaders) {
            if (!loader.exists(name)) {
                continue;
            }
            try {
                return loader.getCacheKey(name);
            }
            catch (e) {
                if (e instanceof loader_1.TwingErrorLoader) {
                    exceptions.push(loader.constructor.name + ': ' + e.message);
                }
            }
        }
        throw new loader_1.TwingErrorLoader(`Template "${name}" is not defined${exceptions.length ? ' (' + exceptions.join(', ') + ')' : ''}.`);
    }
    isFresh(name, time) {
        let exceptions = [];
        for (let loader of this.loaders) {
            if (!loader.exists(name)) {
                continue;
            }
            try {
                return loader.isFresh(name, time);
            }
            catch (e) {
                if (e instanceof loader_1.TwingErrorLoader) {
                    exceptions.push(loader.constructor.name + ': ' + e.message);
                }
            }
        }
        throw new loader_1.TwingErrorLoader(`Template "${name}" is not defined${exceptions.length ? ' (' + exceptions.join(', ') + ')' : ''}.`);
    }
    resolve(name) {
        let exceptions = [];
        for (let loader of this.loaders) {
            if (!loader.exists(name)) {
                continue;
            }
            try {
                return loader.resolve(name);
            }
            catch (e) {
                if (e instanceof loader_1.TwingErrorLoader) {
                    exceptions.push(loader.constructor.name + ': ' + e.message);
                }
            }
        }
        throw new loader_1.TwingErrorLoader(`Template "${name}" is not defined${exceptions.length ? ' (' + exceptions.join(', ') + ')' : ''}.`);
    }
}
exports.TwingLoaderChain = TwingLoaderChain;
