"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class TwingSourceMapNode {
    /**
     * TwingSourceMapTreeNode constructor
     *
     * @param {string} type
     * @param {string} source
     * @param {number} line 0-based
     * @param {number} column 0-based
     */
    constructor(name, source, line, column) {
        this.name = name;
        this.source = source;
        this.originalPosition = {
            line: line,
            column: column
        };
        this.content = null;
        this.parent = null;
        this.children = [];
        this.cursor = {
            line: 0,
            column: 0
        };
    }
    getName() {
        return this.name;
    }
    getSource() {
        return this.source;
    }
    getOriginalPosition() {
        return this.originalPosition;
    }
    getContent() {
        return this.content;
    }
    getParent() {
        return this.parent;
    }
    getChildren() {
        return this.children;
    }
    getCursor() {
        return this.cursor;
    }
    addChild(child) {
        child.parent = this;
        this.children.push(child);
    }
    setContent(content) {
        if (content !== false) {
            this.content = content;
        }
        else {
            this.content = null;
        }
    }
    /**
     * Returns the mappings starting from this node.
     *
     * @returns {Array<Mapping>}
     */
    toMappings() {
        let proceedNode = (node, parentCursor) => {
            let nodeCursor = node.getCursor();
            nodeCursor.line = parentCursor.line;
            nodeCursor.column = parentCursor.column;
            let mappings = [];
            let originalPosition = node.getOriginalPosition();
            mappings.push({
                name: node.getName(),
                source: node.getSource(),
                original: {
                    line: originalPosition.line + 1,
                    column: originalPosition.column
                },
                generated: {
                    line: parentCursor.line + 1,
                    column: parentCursor.column
                }
            });
            for (let child of node.getChildren()) {
                let childMappings = proceedNode(child, node.getCursor());
                for (let mapping of childMappings) {
                    mappings.push(mapping);
                }
            }
            let lines = node.getContent().split('\n');
            let lastLine = lines[lines.length - 1];
            if (lines.length > 1) {
                parentCursor.line += (lines.length - 1);
                parentCursor.column = 0;
            }
            parentCursor.column += lastLine.length;
            return mappings;
        };
        // send a virtual cursor since the parent is out of the scope
        // - i.e we are generating the mappings starting from this node
        return proceedNode(this, {
            line: 0,
            column: 0
        });
    }
}
exports.TwingSourceMapNode = TwingSourceMapNode;
