import { TwingSource } from "./source";
import { TwingEnvironment } from "./environment";
import { TwingTemplateWrapper } from "./template-wrapper";
import { TwingExtensionInterface } from "./extension-interface";
export declare abstract class TwingTemplate {
    static ANY_CALL: string;
    static ARRAY_CALL: string;
    static METHOD_CALL: string;
    protected parent: TwingTemplate | false;
    protected parents: Map<TwingTemplate | string, TwingTemplate>;
    protected env: TwingEnvironment;
    protected blocks: Map<string, Array<any>>;
    protected traits: Map<string, Array<any>>;
    /**
     * @internal
     */
    protected extensions: Map<string, TwingExtensionInterface>;
    constructor(env: TwingEnvironment);
    /**
     * @internal this method will be removed in 2.0 and is only used internally to provide an upgrade path from 1.x to 2.0
     */
    toString(): string;
    /**
     * Returns the template name.
     *
     * @returns {string} The template name
     */
    abstract getTemplateName(): string;
    /**
     * Returns debug information about the template.
     *
     * @returns {Map<number, {line: number, column: number}>} Debug information
     *
     * @internal
     */
    abstract getDebugInfo(): Map<number, {
        line: number;
        column: number;
    }>;
    /**
     * Returns information about the original template source code.
     *
     * @return TwingSource
     */
    getSourceContext(): TwingSource;
    /**
     * Returns the parent template.
     *
     * @param context
     *
     * @returns TwingTemplate|false The parent template or false if there is no parent
     */
    getParent(context?: any): false | TwingTemplate;
    isTraitable(): boolean;
    /**
     * Displays a parent block.
     *
     * This method is for internal use only and should never be called
     * directly.
     *
     * @param {string} name The block name to display from the parent
     * @param context The context
     * @param {Map<string, TwingNodeBlock>} blocks The active set of blocks
     * @returns {string}
     *
     * @internal
     */
    displayParentBlock(name: string, context: any, blocks?: Map<string, Array<any>>): void;
    /**
     * Displays a block.
     *
     * This method is for internal use only and should never be called
     * directly.
     *
     * @param {string} name The block name to display
     * @param context The context
     * @param {Map<string, Array<*>>} blocks The active set of blocks
     * @param {boolean} useBlocks Whether to use the active set of blocks
     *
     * @internal
     */
    displayBlock(name: string, context: any, blocks?: Map<string, Array<any>>, useBlocks?: boolean): void;
    /**
     * Renders a parent block.
     *
     * This method is for internal use only and should never be called
     * directly.
     *
     * @param {string} name The block name to display from the parent
     * @param {*} context The context
     * @param {Map<string, Array<any>>} blocks The active set of blocks
     *
     * @returns string The rendered block
     *
     * @internal
     */
    renderParentBlock(name: string, context: any, blocks?: Map<string, Array<any>>): string;
    /**
     * Renders a block.
     *
     * This method is for internal use only and should never be called
     * directly.
     *
     * @param {string} name The block name to display
     * @param context The context
     * @param {Map<string, Array<*>>} blocks The active set of blocks
     * @param {boolean} useBlocks Whether to use the active set of blocks
     *
     * @return string The rendered block
     *
     * @internal
     */
    renderBlock(name: string, context: any, blocks?: Map<string, [TwingTemplate, string]>, useBlocks?: boolean): string;
    /**
     * Returns whether a block exists or not in the active context of the template.
     *
     * This method checks blocks defined in the active template
     * or defined in "used" traits or defined in parent templates.
     *
     * @param {string} name The block name
     * @param context The context
     * @param {Map<string, Array<*>>} blocks The active set of blocks
     *
     * @returns {boolean} true if the block exists, false otherwise
     */
    hasBlock(name: string, context: any, blocks?: Map<string, Array<any>>): boolean;
    /**
     * Returns all block names in the active context of the template.
     *
     * This method checks blocks defined in the active template
     * or defined in "used" traits or defined in parent templates.
     *
     * @param context The context
     * @param {Map<string, Array<*>>} blocks The active set of blocks
     * @returns {Array<string>}
     */
    getBlockNames(context: any, blocks?: Map<string, Array<any>>): Array<string>;
    loadTemplate(template: TwingTemplate | TwingTemplateWrapper | Array<TwingTemplate> | string, templateName?: string, line?: number, index?: number): TwingTemplate | TwingTemplateWrapper;
    /**
     * Returns all blocks.
     *
     * This method is for internal use only and should never be called
     * directly.
     *
     * @returns {Map<string, Array<*>>} An array of blocks
     *
     * @internal
     */
    getBlocks(): Map<string, any[]>;
    display(context: any, blocks?: Map<string, Array<any>>): void;
    render(context: any): string;
    /**
     * Auto-generated method to display the template with the given context.
     *
     * @param {*} context An array of parameters to pass to the template
     * @param {Map<string, Array<*>>} blocks  An array of blocks to pass to the template
     */
    abstract doDisplay(context: {}, blocks: Map<string, Array<any>>): void;
    protected doGetParent(context: any): TwingTemplate | string | false;
    protected displayWithErrorHandling(context: any, blocks?: Map<string, Array<any>>): void;
}
